/* databaseProgram.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "databaseProgram.h"
#include "tableInfo.h"
#include "postgresDatabase.h"
#include "mySQLDatabase.h"

#include <getopt.h>

static const string default_dbname = "paloma";

DatabaseProgram::
DatabaseProgram(const string &program_name) : MainProgram(program_name) {
  define_opt("db", true);
  define_opt("dbhost", true);
  define_opt("dbport", true);
  define_opt("dbuser", true);
  define_opt("dbpasswd", true);
#ifdef USE_PGSQL
  define_opt("pgsql", true);
#endif
#ifdef USE_MYSQL
  define_opt("mysql", true);
#endif

  // Default parameter values.
  _dbtype = DBT_unknown;
  _dbname = default_dbname;
  _dbport = 0;
}

void DatabaseProgram::
command_line(int argc, char *argv[]) {
  MainProgram::command_line(argc, argv);
}

void DatabaseProgram::
init() {

#ifdef USE_PGSQL
  if (_dbtype == DBT_unknown || _dbtype == DBT_pgsql) {
    music_db = new PostgresDatabase(_dbname, _dbhost, _dbport, 
				    _dbuser, _dbpasswd);
  } else
#endif
#ifdef USE_MYSQL
  if (_dbtype == DBT_unknown || _dbtype == DBT_mysql) {
    music_db = new MySQLDatabase(_dbname, _dbhost, _dbport, 
				 _dbuser, _dbpasswd);
  } else
#endif
  {
    cerr << "No database engine specified!\n";
    exit(1);
  }
  
  MainProgram::init();
}

void DatabaseProgram::
handle_opts(const string &opt, const string &arg) {
  if (opt == "db") {
    _dbname = arg;

  } else if (opt == "dbhost") {
    _dbhost = arg;

  } else if (opt == "dbport") {
    _dbport = atoi(arg.c_str());

  } else if (opt == "dbuser") {
    _dbuser = arg;

  } else if (opt == "dbpasswd") {
    _dbpasswd = arg;

#ifdef USE_PGSQL
  } else if (opt == "pgsql") {
    PostgresDatabase::extra_args(arg);
    if (_dbtype == DBT_unknown) {
      _dbtype = DBT_pgsql;
    }
#endif

#ifdef USE_MYSQL
  } else if (opt == "mysql") {
    MySQLDatabase::extra_args(arg);
    if (_dbtype == DBT_unknown) {
      _dbtype = DBT_mysql;
    }
#endif

  } else {
    MainProgram::handle_opts(opt, arg);
  }
}

void DatabaseProgram::
describe_options() {
  cerr <<
    "  -db name   Connect to the named database.  The default is '" 
       << default_dbname << "'.\n"
    "  -dbhost host   Contact the SQL database on the indicated host, if the SQL\n"
    "             backend supports it.\n"
    "  -dbport port   Use the indicated port number to contact the SQL database.\n"
    "  -dbuser user   Specify the name of the user to access the SQL database as.\n"
    "             This is not necessarily related to the Unix system users; it\n"
    "             is defined by the SQL backend.\n"
    "  -dbpasswd pwd  Give a password for the SQL user, above.\n"

#ifdef USE_PGSQL
    "  -pgsql 'args'  Specify that PostgreSQL should be used as the default\n"
    "             database engine, and also specify optional arguments that\n"
    "             have meaning specific to Postgres.\n"
#endif
#ifdef USE_MYSQL
    "  -mysql 'args'  Specify that MySQL should be used as the default\n"
    "             database engine, and also specify optional arguments that\n"
    "             have meaning specific to MySQL.\n"
#endif
    ;
  MainProgram::describe_options();
}
