/* referenceCount.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"
#include "referenceCount.h"

#include <assert.h>
#include <stdlib.h>

// An unlikely-to-occur random number.  This is to be assigned to
// ReferenceCount::_alive_test only while the ReferenceCount object
// has not been deleted.
static const int alive_test_check = 0x3423413;      

ReferenceCount::
ReferenceCount() {
  // The initial reference count is created at zero.  This is in
  // contrast to GTK's convention of creating it at one.  That means
  // that the count should be referenced the first time its pointer is
  // assigned.
  _ref_count = 0;

#ifndef NDEBUG
  _alive_test = alive_test_check;
#endif
}

ReferenceCount::
ReferenceCount(const ReferenceCount &copy) {
  // The copy constructor must not copy the reference count too!  We
  // have our own reference count.
  _ref_count = 0;

#ifndef NDEBUG
  assert(_alive_test == alive_test_check);
  _alive_test = 0;
#endif
}

ReferenceCount &ReferenceCount::
operator = (const ReferenceCount &copy) {
  // Similarly, the assignment operator must not change the reference
  // count.
  return *this;
}

ReferenceCount::
~ReferenceCount() {
  assert(_ref_count == 0);
}


ReferenceCountProxy::
ReferenceCountProxy(ReferenceCount *ptr) {
  _ptr = NULL;
  set_ptr(ptr);
}

ReferenceCountProxy::
ReferenceCountProxy(const ReferenceCountProxy &copy) {
  _ptr = NULL;
  set_ptr(copy._ptr);
}

ReferenceCountProxy &ReferenceCountProxy::
operator = (const ReferenceCountProxy &copy) {
  set_ptr(copy._ptr);
  return *this;
}

ReferenceCountProxy::
~ReferenceCountProxy() {
  set_ptr(NULL);
}

bool ReferenceCountProxy::
operator == (const ReferenceCountProxy &other) const {
  return _ptr == other._ptr;
}

void ReferenceCountProxy::
clear() {
  set_ptr(NULL);
}


void ReferenceCountProxy::
set_ptr(ReferenceCount *ptr) {
  // Don't do anything if we're assigning the pointer to the same
  // value it used to be.  We don't want to accidentally delete the
  // object by dropping its ref count and immediately incrementing it.
  if (_ptr != ptr) {

    // If the previous value of the pointer was non-NULL, decrement
    // its ref count and possibly free it.
    if (_ptr != NULL) {
      assert(_ptr->_ref_count > 0);
      _ptr->_ref_count--;
      if (_ptr->_ref_count == 0) {
	delete _ptr;
      }
    }

    // If the new value is non-NULL, increment its ref count.
    _ptr = ptr;
    if (_ptr != NULL) {
      _ptr->_ref_count++;
      assert(_ptr->_ref_count > 0);
    }
  }
}

ReferenceCount *ReferenceCountProxy::
get_ptr() const {
  assert(_ptr == NULL || _ptr->_alive_test == alive_test_check);
  return _ptr;
}

