/* commandInterface.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "commandInterface.h"
#include "player.h"

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>

#include <map>

// We need to track all the pids that we might expect to get a SIGCHLD
// from.
typedef map<pid_t, CommandInterface *> WaitingPids;
static WaitingPids _waiting_pids;

static bool sigchld_set = false;
static void sigchld_handler(int signo) {
  // Some child just terminated.  Figure out which one.
  pid_t pid = waitpid(-1, NULL, WNOHANG);
  if (pid > 0) {
    WaitingPids::iterator wpi = _waiting_pids.find(pid);
    if (wpi != _waiting_pids.end()) {
      CommandInterface *interface = (*wpi).second;

      // The CommandInterface pointer associated with the given child
      // pid might be NULL, indicating it's already destructed.
      // That's ok.  Or, it might be non-NULL, in which case the
      // exiting child is a surprise, and we need to let it know.

      if (interface == NULL || interface->child_exited(pid)) {
	_waiting_pids.erase(wpi);
      }

    } else {
      cerr << "Unexpected child " << pid << " terminated.\n";
    }
  }
}

CommandInterface::
CommandInterface(const string &command,
		 const PlayerDef::Status &status,
		 Player *player) :
  _status(status),
  _player(player)
{
  if (!sigchld_set) {
    // Set up the interrupt handler so we'll note when our child
    // process goes away.
    struct sigaction act;
    act.sa_handler = sigchld_handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP;
    if (sigaction(SIGCHLD, &act, NULL) < 0) {
      perror("sigaction");
      exit(1);
    }

    sigchld_set = true;
  }

  _pid = 0;
  _fd = -1;
  _anything_played = false;
  _total_frames = _player->get_total_frames();
  _total_time = _player->get_total_time();
  _current_position = 0.0;

  int pipe_fd[2];
  if (pipe(pipe_fd) < 0) {
    perror("pipe");
    return;
  }

  cerr << command << "\n";
  if (_player->get_verbose()) {
    _player->feedback_user("send: " + command);
  }

  // We must block the SIGCHLD signal until we've fully set up the
  // child.  Otherwise, we could miss the signal (if the child
  // terminates immediately), or worse, enter a race condition with
  // updates to the _waiting_pids map.

  sigset_t sigchld_set;
  sigemptyset(&sigchld_set);
  sigaddset(&sigchld_set, SIGCHLD);
  sigprocmask(SIG_BLOCK, &sigchld_set, NULL);

  _pid = fork();
  if (_pid < 0) {
    perror("fork");
    return;
  }

  if (_pid == 0) {
    // Child.

    sigprocmask(SIG_UNBLOCK, &sigchld_set, NULL);

    // We want to make this its own process group, so that we'll be
    // able to easily kill all of the spawned processes.
    setpgid(0, 0);

    // Standard output and standard error go to the pipe.
    dup2(pipe_fd[1], 1);
    dup2(pipe_fd[1], 2);

    execl("/bin/sh", "/bin/sh", "-c", command.c_str(), NULL);
    perror("execlp");
    cerr << "Unable to execute " << command << "\n";
    _exit(1);
  }

  // Parent.
  _waiting_pids[_pid] = this;

  // Now it's safe to catch the SIGCHLD signal again.
  sigprocmask(SIG_UNBLOCK, &sigchld_set, NULL);

  _fd = pipe_fd[0];

  if (_status.empty()) {
    // If we have no status checking ability, we have to assume it
    // played something.
    _anything_played = true;
  }
}

CommandInterface::
~CommandInterface() {
  if (_pid > 0) {
    // If we still have a pid when we destruct, we need to remove it
    // from the _waiting_pids map, so that the SIGCHLD handler won't
    // end up with a bogus CommandInterface pointer when our child
    // does eventually terminate.  But first we need to temporarily
    // block SIGCHLD while we fiddle with the map.

    sigset_t sigchld_set;
    sigemptyset(&sigchld_set);
    sigaddset(&sigchld_set, SIGCHLD);
    sigprocmask(SIG_BLOCK, &sigchld_set, NULL);

    // Instead of removing the pid altogether, we'll just set its
    // associated pointer to NULL.  That way the SIGCHLD handler will
    // know that this pid was associated with a recently-deleted
    // CommandInterface object.
    _waiting_pids[_pid] = NULL;

    sigprocmask(SIG_UNBLOCK, &sigchld_set, NULL);

    // And now it's safe to kill the child.
    kill(-_pid, SIGTERM);
  }
}

int CommandInterface::
get_fd() const {
  return _fd;
}

bool CommandInterface::
read_status() {
  if (_fd < 0) {
    return false;
  }

  char buffer[4096];
  int bytes_read = read(_fd, buffer, 4096);
  if (bytes_read > 0) {
    // Look for the indicated status string.
    line_up_process_status(string(buffer, bytes_read));
    
    return true;
  }

  // We might have closed and reset the file descriptor in the sigchld
  // interrupt handler, if our child process terminated.  In that
  // case, _fd was valid when we entered the function, but might be -1
  // now.  We only want to return true if _fd is still valid.
  return (_fd >= 0);
}

bool CommandInterface::
child_exited(pid_t pid) {
  if (pid != _pid) {
    cerr << "Unexpected child_exited() call for pid " << pid 
	 << " (expected " << _pid << ")\n";
    return false;
  }

  close(_fd);
  _fd = -1;
  _pid = 0;
  return true;
}

bool CommandInterface::
anything_played() const {
  return _anything_played;
}

double CommandInterface::
get_current_time() const {
  return _current_position * _total_time;
}

int CommandInterface::
get_total_frames() const {
  return _total_frames;
}
 
double CommandInterface::
get_total_time() const {
  return _total_time;
}

// Accept a bunch of output from the child process, which might be cut
// in arbitrary places; reassemble it into lines and pass it
// line-by-line into match_status().
void CommandInterface::
line_up_process_status(const string &input) {
  _last_input += input;

  int sol = 0;
  int eol = _last_input.find_first_of("\n\r", sol);
  while (eol != string::npos) {
    match_status(_last_input.substr(sol, eol - sol));
    sol = eol + 1;
    eol = _last_input.find_first_of("\n\r", sol);
  }

  _last_input = _last_input.substr(sol);

  if (_pid == 0 && !_last_input.empty()) {
    // If this is the last bit of output from the process, send it
    // through even if it doesn't have a trailing newline.
    match_status(_last_input);
    _last_input = string();
  }
}

void CommandInterface::
match_literal(const string &line, const string &literal, int &next_pos) {
  next_pos = line.find(literal, next_pos);

  if (next_pos != string::npos) {
    // We matched the literal; set the next position to the end of the
    // literal.
    next_pos += literal.length();
  }
}

void CommandInterface::
extract_integer(const string &line, int &next_pos, int &result) {
  int pos = next_pos;
  while (pos < line.length() && isspace(line[pos])) {
    pos++;
  }
  int start_digits = pos;
  while (pos < line.length() && isdigit(line[pos])) {
    pos++;
  }

  if (pos == start_digits) {
    // There wasn't a number there!  Flag it as an error.
    next_pos = string::npos;
  } else {
    result = atoi(line.substr(next_pos, pos - next_pos).c_str());
    next_pos = pos;
  }
}

void CommandInterface::
extract_float(const string &line, int &next_pos, double &result) {
  int pos = next_pos;
  while (pos < line.length() && isspace(line[pos])) {
    pos++;
  }
  int start_digits = pos;
  while (pos < line.length() && isdigit(line[pos])) {
    pos++;
  }
  if (pos < line.length() && line[pos] == '.') {
    // A decimal point!
    pos++;
    while (pos < line.length() && isdigit(line[pos])) {
      pos++;
    }
  }

  if (pos == start_digits) {
    // There wasn't a number there!  Flag it as an error.
    next_pos = string::npos;
  } else {
    result = atof(line.substr(next_pos, pos - next_pos).c_str());
    next_pos = pos;
  }
}

void CommandInterface::
extract_time(const string &line, int &next_pos, double &result) {
  // Look for [[HH:]MM:]SS[.frac]
  int secs;
  extract_integer(line, next_pos, secs);
  if (next_pos != string::npos) {
    // We got the first integer.  Hours, minutes or seconds?
    if (next_pos < line.length() && line[next_pos] == ':') {
      // That wasn't seconds, that was minutes.  Get the seconds.
      next_pos++;
      int mins = secs;

      extract_integer(line, next_pos, secs);
      if (next_pos != string::npos) {
	// We got the second integer.  Minutes or seconds?
	if (next_pos < line.length() && line[next_pos] == ':') {
	  // Huh.  No, THAT was minutes.  Get the seconds.
	  next_pos++;
	  int hours = mins;
	  mins = secs;
	  extract_integer(line, next_pos, secs);
	  mins += 60*hours;
	}
      }
      secs += 60*mins;
    }

    // And is there a fractional part?
    double frac = 0.0;
    if (next_pos < line.length() && line[next_pos] == '.') {
      int pos = next_pos+1;
      while (pos < line.length() && isdigit(line[pos])) {
	pos++;
      }
      frac = atof(line.substr(next_pos, pos - next_pos).c_str());
      next_pos = pos;
    }

    if (next_pos != string::npos) {
      result = (double)secs + frac;
    }
  }
}

void CommandInterface::
match_status(const string &line) {
  if (_player->get_verbose()) {
    _player->feedback_user("read: " + line);
  }

  // Parse the string output by the sub-process for information about
  // where we are in the sound file.
  int current_frame = -1;
  int frames_remaining = -1;
  int total_frames = -1;

  double current_time = -1.0;
  double time_remaining = -1.0;
  double total_time = -1.0;

  double current_position = -1.0;

  PlayerDef::Status::const_iterator si;
  si = _status.begin();
  int next_pos = 0;
  while (si != _status.end() && next_pos != string::npos) {
    switch ((*si)._type) {
    case PlayerDef::ST_literal:
      match_literal(line, (*si)._literal, next_pos);
      break;

    case PlayerDef::ST_current_frame:
      extract_integer(line, next_pos, current_frame);
      break;

    case PlayerDef::ST_frames_remaining:
      extract_integer(line, next_pos, frames_remaining);
      break;

    case PlayerDef::ST_total_frames:
      extract_integer(line, next_pos, total_frames);
      break;

    case PlayerDef::ST_current_time:
      extract_time(line, next_pos, current_time);
      break;

    case PlayerDef::ST_time_remaining:
      extract_time(line, next_pos, time_remaining);
      break;

    case PlayerDef::ST_total_time:
      extract_time(line, next_pos, total_time);
      break;

    case PlayerDef::ST_current_pos:
      extract_float(line, next_pos, current_position);
      break;

    default:
      cerr << "Unexpected PlayerDef::StatusType " 
	   << (*si)._type << "\n";
      abort();
    }

    ++si;
  }

  if (next_pos != string::npos) {
    // If we got through the end of the status match list and matched
    // everything, hooray!  It all counts.
    if (total_time >= 0) {
      _total_time = total_time;
    }
    if (current_time >= 0) {
      if (time_remaining >= 0) {
	_total_time = current_time + time_remaining;
      }
      _current_position = (double)current_time / (double)_total_time;
    }

    if (total_frames >= 0) {
      _total_frames = total_frames;
    }
    if (current_frame >= 0) {
      if (frames_remaining >= 0) {
	_total_frames = current_frame + frames_remaining;
      }
      _current_position = (double)current_frame / (double)_total_frames;
    }

    if (current_position >= 0.0) {
      _current_position = current_position;
    }

    _anything_played = true;
  }
}

